<?php

namespace Drupal\sqrl;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException;
use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Component\Utility\Random;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityTypeManager;
use Drupal\Core\Session\AccountProxy;
use Drupal\sqrl\Entity\Identity;
use Drupal\sqrl\Exception\ClientException;
use Drupal\user\Entity\User;
use Exception;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Client service.
 */
class Client {

  use StringManipulation;

  const CRLF = "\r\n";
  const VERSION = '1';

  const FLAG_IDK_MATCH = 0x01;
  const FLAG_PIDK_MATCH = 0x02;
  const FLAG_IP_MATCH = 0x04;
  const FLAG_ACCOUNT_DISABLED = 0x08;
  const FLAG_ACCOUNT_LOGGED_IN = 0x10;
  const FLAG_ACCOUNT_CREATION_ALLOWED = 0x20;
  const FLAG_COMMAND_FAILURE = 0x40;
  const FLAG_FAILURE = 0x80;
  const FLAG_STALE_NUT = 0x100;

  /**
   * @var \Drupal\Core\Entity\EntityTypeManager
   */
  protected $entityTypeManager;

  /**
   * @var \Drupal\sqrl\State
   */
  protected $state;

  /**
   * @var \Symfony\Component\HttpFoundation\Request
   */
  protected $request;

  /**
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $config;

  /**
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $siteSettings;

  /**
   * @var \Drupal\Core\Session\AccountProxy
   */
  protected $currentUser;

  /**
   * @var \Drupal\sqrl\Sqrl
   */
  protected $sqrl;

  /**
   * @var \Drupal\sqrl\Log
   */
  protected $log;

  /**
   * @var \Drupal\Component\Utility\Random
   */
  protected $random;

  /**
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $time;

  /**
   * @var \Drupal\sqrl\SqrlActionPluginManager
   */
  protected $actionPluginManager;

  /**
   * @var \Drupal\sqrl\Nut
   */
  private $nut;

  private $postValues;
  private $clientSignatures;
  private $clientHeader;
  private $clientVars;
  private $serverVars;
  private $validationString;
  private $http_code = 403;
  private $message = '';
  private $tif = 0;
  private $response = [];
  private $fields = [];
  private $account;

  /**
   * Constructs a FormAlter object.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   * @param \Drupal\Core\Session\AccountProxy $current_user
   * @param \Drupal\sqrl\Sqrl $sqrl
   * @param \Drupal\sqrl\Log $log
   * @param \Drupal\Component\Datetime\TimeInterface $time
   * @param \Drupal\sqrl\SqrlActionPluginManager $sqrl_action_plugin_manager
   * @param \Drupal\sqrl\State $state
   * @param \Symfony\Component\HttpFoundation\RequestStack $request
   * @param \Drupal\Core\Entity\EntityTypeManager $entity_type_manager
   */
  public function __construct(ConfigFactoryInterface $config_factory, AccountProxy $current_user, Sqrl $sqrl, Log $log, TimeInterface $time, SqrlActionPluginManager $sqrl_action_plugin_manager, State $state, RequestStack $request, EntityTypeManager $entity_type_manager) {
    $this->config = $config_factory->get('sqrl.settings');
    $this->siteSettings = $config_factory->get('system.site');
    $this->currentUser = $current_user;
    $this->sqrl = $sqrl;
    $this->log = $log;
    $this->actionPluginManager = $sqrl_action_plugin_manager;
    $this->time = $time;
    $this->state = $state;
    $this->entityTypeManager = $entity_type_manager;

    $this->request = $request->getCurrentRequest();
    $this->random = new Random();

    $this->nut = $sqrl->getNewNut();
    $this->nut->fetch();
  }

  /**
   * @return \Drupal\sqrl\Nut
   */
  public function getNut(): Nut {
    return $this->nut;
  }

  /**
   * @return int
   */
  public function getHttpCode(): int {
    return $this->http_code;
  }

  /**
   * @return int
   */
  public function getTif(): int {
    return $this->tif;
  }

  /**
   * @param string $key
   *
   * @return string
   */
  public function getClientVar($key): string {
    return $this->clientVars[$key] ?? '';
  }

  /**
   * @param string $key
   *
   * @return string
   */
  public function getClientOpt($key): string {
    $opts = explode('~', $this->getClientVar('opt'));
    return isset($opts[$key]);
  }

  /**
   * @param string $key
   * @param string $value
   *
   * @return \Drupal\sqrl\Client
   */
  public function setResponse($key, $value): Client {
    $this->response[$key] = $value;
    return $this;
  }

  /**
   * @param string $msg
   *
   * @return $this
   */
  public function setMessage($msg): self {
    $this->message = $msg;
    return $this;
  }

  /**
   * https://www.grc.com/sqrl/semantics.htm: The data to be signed are the two
   * base64url encoded values of the “client=” and “server=” parameters with the
   * “server=” value concatenated to the end of the “client=” value.
   *
   * @return string
   */
  public function process(): string {
    #SQRL::get_message()->log(SQRL_LOG_LEVEL_DEBUG, 'Incoming client request');

    $this->clientSignatures = [
      'ids' => $this->getPostValue('ids'),
      // 'pids' => $this->getPostValue('pids'),
      // 'urs' => $this->getPostValue('urs'),
    ];
    $this->clientHeader = [
      'host' => $this->getServerValue('HTTP_HOST'),
      'auth' => $this->getServerValue('HTTP_AUTHENTICATION'),
      'agent' => $this->getServerValue('HTTP_USER_AGENT'),
    ];
    $this->clientVars = $this->decodeParameter($this->getPostValue('client'));
    $this->serverVars = $this->decodeParameter($this->getPostValue('server'));
    $this->validationString = $this->getPostValue('client', TRUE) . $this->getPostValue('server', TRUE);

    try {
      $this->validate();
    }
    catch (ClientException $e) {
      $this->tif |= self::FLAG_COMMAND_FAILURE;
      $this->message = $e->getMessage();
      $this->log->error($e->getMessage());
      return '';
    }

    try {
      $this->validateIpAddress();
      $this->authenticate();
    }
    catch (ClientException $e) {
      $this->tif |= self::FLAG_COMMAND_FAILURE;
      $this->message = $e->getMessage();
      $this->log->error($e->getMessage());
      return '';
    }

    $this->http_code = 200;
    $this->message = 'Welcome';

    $this->executeCommands(explode('~', $this->clientVars['cmd']));

    // Build the response body.
    $response = [
      'ver' => self::VERSION,
      'nut' => $this->sqrl->getNut()->getPublicNut(),
      'tif' => $this->tif,
      'qry' => $this->sqrl->getPath('sqrl.client', [], TRUE, FALSE),
      'sfn' => $this->siteSettings->get('name'),
    ];
    foreach ($this->response as $key => $value) {
      $response[$key] = $value;
    }

    return $this->encodeResponse($response, TRUE);
  }

  /**
   * @throws \Drupal\sqrl\Exception\ClientException
   */
  private function validate() {
    $this->validateSignatures();
    $this->validateHeader();
    $this->validateClientVars();
    $this->validateServerVars();
    $this->validateNut();
  }

  /**
   * @throws \Drupal\sqrl\Exception\ClientException
   */
  private function validateSignatures() {
    $msg = $this->validationString;
    // Is the message present?
    if (empty($msg)) {
      throw new ClientException('Missing validation string');
    }
    $required_signatures = [
      'ids' => [
        'key' => 'idk',
        'required' => TRUE,
      ],
      'pids' => [
        'key' => 'pidk',
        'required' => FALSE,
      ],
      'urs' => [
        'key' => 'purs',
        'required' => FALSE,
      ],
    ];
    foreach ($required_signatures as $sig_key => $data) {
      $sig = $this->clientSignatures[$sig_key] ?? '';
      $pk  = $this->clientVars[$data['key']] ?? '';
      if ($data['required'] && (empty($sig) || empty($pk))) {
        // Is the signature present?
        if (empty($sig)) {
          throw new ClientException('Missing signature');
        }
        // Is the public key present?
        if (empty($pk)) {
          throw new ClientException('Missing public key');
        }
      }
      if (!empty($sig) && !empty($pk)) {
        // Validate signature
        $this->validateSignature($msg, $sig, $pk);
      }
    }
  }

  /**
   * Validate a specific signature.
   *
   * @param string $msg
   *  plain text of signed message
   * @param string $sig
   *  base64url-encoded signature
   * @param string $pk
   *  base64url-encoded Public Key
   *
   * @throws \Drupal\sqrl\Exception\ClientException
   */
  private function validateSignature($msg, $sig, $pk) {
    if (!sodium_crypto_sign_verify_detached($this->base64_decode($sig), $msg, $this->base64_decode($pk))) {
      throw new ClientException('Invalid signature');
    }
  }

  /**
   * Validate required header values.
   */
  private function validateHeader() {
    // No validation required at this point.
  }

  /**
   * Validate required client values.
   *
   * @throws \Drupal\sqrl\Exception\ClientException
   */
  private function validateClientVars() {
    if ($this->clientVars['ver'] !== self::VERSION) {
      throw new ClientException('Unsupported version');
    }
  }

  /**
   * Validate required server values.
   */
  private function validateServerVars() {
    // No validation required at this point.
  }

  /**
   * Validate nut.
   *
   * @throws \Drupal\sqrl\Exception\ClientException
   */
  private function validateNut() {
    if (!$this->getNut()->isValid()) {
      if ($this->getNut()->isExpired()) {
        $this->tif |= self::FLAG_STALE_NUT;
      }
      throw new ClientException('Invalid nut');
    }

    // TODO: Check if this client request is a response to a previous one and
    //   validate extra arguments.
  }

  /**
   * Validate IP address.
   *
   * @throws \Drupal\sqrl\Exception\ClientException
   */
  private function validateIpAddress() {
    if ($this->getNut()->isIpValid()) {
      $this->tif |= self::FLAG_IP_MATCH;
      return;
    }
    if ($this->getClientOpt('noiptest')) {
      return;
    }
    throw new ClientException('Invalid nut');
  }

  /**
   * @param array $values
   * @param bool $encode
   *
   * @return string
   */
  private function encodeResponse($values, $encode): string {
    $output = array();
    foreach ($values as $key => $value) {
      $output[] = $key . '=' . $value;
    }
    if ($encode) {
      return $this->base64_encode(implode(self::CRLF, $output) . self::CRLF);
    }
    return implode(self::CRLF, $output);
  }

  /**
   * @param string $key
   *
   * @param bool $plain
   *
   * @return string
   */
  private function getPostValue($key, $plain = FALSE): string {
    if (!isset($this->postValues)) {
      $this->postValues = [];
      foreach (explode('&', $this->request->getContent()) as $entry) {
        list($key, $value) = explode('=', $entry);
        $this->postValues[$key] = ($key === 'ids') ? $value : $this->base64_decode($value);
      }
      $this->log->debug('POST %values', ['%values' => json_encode($this->postValues, JSON_PRETTY_PRINT)]);
    }
    $value = $this->postValues[$key] ?? '';
    if ($plain) {
      return $this->base64_encode($value);
    }
    return $value;
  }

  /**
   * Get the value of a key in the $_SERVER array or the string unknown if the
   * key is not set.
   *
   * @param string $key
   *
   * @return string
   */
  private function getServerValue($key): string {
    return $_SERVER[$key] ?? 'unknown';
  }

  /**
   * @param string $param
   *
   * @return array
   */
  private function decodeParameter($param): array {
    $values = explode(self::CRLF, $param);
    $vars = [];
    foreach ($values as $value) {
      if (!empty($value)) {
        $parts = explode('=', $value);
        $k = array_shift($parts);
        $vars[$k] = implode('=', $parts);
      }
    }
    return $vars;
  }

  /**
   * @param string $key_type
   *   Either "idk" or "pidk".
   *
   * @return \Drupal\user\Entity\User|null
   */
  private function findUserAccountByKey($key_type) {
    if (!empty($this->clientVars[$key_type])) {
      return $this->findUserAccount($this->clientVars[$key_type]);
    }
    return NULL;
  }

  /**
   * @param string $key
   *
   * @return \Drupal\user\Entity\User|null
   */
  private function findUserAccount($key) {
    // Make sure we find the enabled accounts first (order by status descending)
    // just in case that there are multiple accounts for the same ID.
    try {
      $id = $this->entityTypeManager->getStorage('sqrl_identity')->getQuery()
        ->condition('idk', $key)
        ->execute();
    } catch (InvalidPluginDefinitionException $e) {
    } catch (PluginNotFoundException $e) {
    }
    if (isset($id) && $id && $identity = Identity::load(reset($id))) {
      return User::load($identity->getUid());
    }
    return NULL;
  }

  /**
   * @throws ClientException
   */
  private function authenticate() {
    $current_account = $this->findUserAccountByKey('idk');
    $previous_account = $this->findUserAccountByKey('pidk');

    if ($previous_account !== NULL) {
      $this->tif |= self::FLAG_PIDK_MATCH;
      $account = $previous_account;
    }
    if ($current_account !== NULL) {
      $this->tif |= self::FLAG_IDK_MATCH;
      $account = $current_account;
      /** @noinspection NullPointerExceptionInspection */
      if ($previous_account && $previous_account->id() !== $account->id()) {
        throw new ClientException('Current and previous user accounts do not match');
      }
    }

    if (!isset($account)) {
      return;
    }

    if (!$account->isActive()) {
      $this->tif |= self::FLAG_ACCOUNT_DISABLED;
      throw new ClientException('Account disabled');
    }

    /*
     * TODO: Let's see if we have to implement this.
    if ($account->isAuthenticated()) {
      $this->tif |= self::FLAG_ACCOUNT_LOGGED_IN;
    }*/

    $this->account = $account;
  }

  /**
   * @param array $commands
   */
  private function executeCommands($commands) {
    $this->log->debug('Execute START: ' . $this->getNut()->getPublicNut());
    $this->log->debug('Client operation: ' . $this->getNut()->getClientOperation());
    try {
      foreach ($commands as $command) {
        $this->log->debug('Execute ' . $command);
        if ($command === 'ident') {
          $command = 'login';
        }
        /** @var \Drupal\sqrl\SqrlActionInterface $plugin */
        $plugin = $this->actionPluginManager->createInstance($command);
        $plugin->setUser($this->account);
        if ($plugin->run()) {
          $this->log->debug('Command succesful');
          $plugin->rememberSuccess();
        }
      }
    }
    catch (Exception $e) {
      $this->tif |= self::FLAG_COMMAND_FAILURE;

      // Message to be returned to the client.
      $this->message = $e->getMessage();

      // Message to be displayed in the browser.
      $this->log->error($e->getMessage());
    }
  }

}
