<?php

namespace Drupal\sqrl;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Session\AccountInterface;
use Drupal\sqrl\Exception\NutException;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Nut service.
 */
final class Nut implements ContainerInjectionInterface {

  use StringManipulation;

  const MODE_BUILD = 'build';
  const MODE_FETCH = 'fetch';
  const IS_COOKIE  = FALSE;

  const STATUS_INITED  = 'inited';
  const STATUS_INVALID = 'invalid';
  const STATUS_BUILT   = 'built';
  const STATUS_FETCHED = 'fetched';

  private $mode = self::MODE_BUILD;
  private $status = self::STATUS_INITED;

  private $expired;

  protected $nut_public;
  protected $nut_encoded;
  protected $nut_raw;

  protected $clientOperation;
  protected $clientOperationParams = [];
  protected $clientMessages = [];
  protected $clientUid;

  /**
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  protected $time;

  /**
   * @var \Symfony\Component\HttpFoundation\Request
   */
  protected $request;

  /**
   * @var \Drupal\sqrl\State
   */
  protected $state;

  /**
   * @var \Drupal\sqrl\Sqrl
   */
  protected $sqrl;

  /**
   * @var \Drupal\sqrl\Log
   */
  protected $log;

  /**
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * Constructs a FormAlter object.
   *
   * @param \Drupal\Core\Session\AccountInterface $current_user
   * @param \Drupal\Component\Datetime\TimeInterface $time
   * @param \Symfony\Component\HttpFoundation\RequestStack $request
   * @param \Drupal\sqrl\State $state
   * @param \Drupal\sqrl\Sqrl $sqrl
   * @param \Drupal\sqrl\Log $log
   */
  public function __construct(AccountInterface $current_user, TimeInterface $time, RequestStack $request, State $state, Sqrl $sqrl, Log $log) {
    $this->currentUser = $current_user;
    $this->time = $time;
    $this->request = $request->getCurrentRequest();
    $this->state = $state;
    $this->sqrl = $sqrl;
    $this->log = $log;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('current_user'),
      $container->get('datetime.time'),
      $container->get('request_stack'),
      $container->get('sqrl.state'),
      $container->get('sqrl.handler'),
      $container->get('sqrl.log')
    );
  }

  /**
   * @param string $op
   *
   * @return \Drupal\sqrl\Nut
   */
  public function setClientOperation($op): Nut {
    $this->clientOperation = $op;
    return $this;
  }

  /**
   * @return string
   */
  public function getClientOperation(): string {
    return $this->clientOperation;
  }

  /**
   * @return string
   */
  public function getClientUid(): string {
    return $this->clientUid;
  }

  /**
   * @return string
   */
  public function getPublicNut(): string {
    $this->build();
    return $this->nut_public;
  }

  /**
   * @return bool
   */
  public function isValid(): bool {
    return ($this->status !== self::STATUS_INVALID);
  }

  /**
   * @return bool
   */
  public function isExpired(): bool {
    return $this->expired ?? FALSE;
  }

  /**
   * @return bool
   */
  public function isIpValid(): bool {
    return ($this->nut_raw['ip'] === $this->request->getClientIp());
  }

  /**
   *
   */
  private function encrypt() {
    $data = $this->base64_encode($this->nut_encoded);
    $this->nut_public = strtr($data, array('+' => '-', '/' => '_', '=' => ''));
  }

  /**
   *
   * @throws \Drupal\sqrl\Exception\NutException
   */
  private function encode() {
    $ref = $this->nut_raw;
    //format bytes
    $this->nut_encoded = pack('L', $ref['time']) .
      pack('L', $ref['counter']) .
      $ref['random'] .
      $this->_dtr_pton($ref['ip']);
  }

  /**
   * @return array
   */
  private function getParams(): array {
    return array(
      'op' => $this->getClientOperation(),
      'ip' => $this->request->getClientIp(),
      'params' => $this->clientOperationParams,
      'messages to browser' => $this->clientMessages,
      'uid' => $this->currentUser->id(),
    );
  }

  /**
   *
   */
  private function build() {
    if ($this->mode !== self::MODE_BUILD || $this->status === self::STATUS_BUILT) {
      return;
    }
    try {
      $this->status = self::STATUS_BUILT;
      $this->nut_raw = array(
        'time' => $this->time->getRequestTime(),
        'ip' => $this->request->getClientIp(),
        'counter' => $this->counter(),
        'random' => $this->randomBytes(4),
      );
      $this->encode();
      $this->encrypt();

      $this->state->setNut($this->nut_public, $this->getParams());

      #setcookie('sqrl', $this->nut_public, $this->time->getRequestTime() + 600, $this->sqrl->getNutBasePath(), $this->sqrl->getNutDomain());
    }
    catch (NutException $e) {
      // TODO: Logging.
      $this->status = self::STATUS_INVALID;
      #$this->error_code = $e->getCode();
      #$this->error_message = $e->getMessage();
    }
  }

  /**
   * @return string
   */
  public function __toString() {
    return $this->getPublicNut();
  }

  /**
   *
   */
  public function fetch() {
    $this->mode = self::MODE_FETCH;
    if ($this->status === self::STATUS_FETCHED) {
      return;
    }
    try {
      $this->nut_public = $this->fetchNut();
      $this->decrypt();
      $this->decode();
      $this->validate_expiration();
      $this->load();
      $this->status = self::STATUS_FETCHED;
    }
    catch (NutException $e) {
      // TODO: Logging.
      $this->status = self::STATUS_INVALID;
      # $this->error_code = $e->getCode();
      # $this->error_message = $e->getMessage();
      $this->log->debug('Fetch NUT error: ' . $e->getMessage());
    }
  }

  /**
   * @return string
   *
   * @throws NutException
   */
  private function fetchNut(): string {
    $nut = $this->request->query->get('nut');
    if (!$nut) {
      throw new NutException('Nut missing from GET request');
    }
    return $nut;
  }

  /**
   *
   */
  private function decrypt() {
    $ref = $this->nut_public;
    $this->nut_encoded = $this->base64_decode(strtr($ref, array('-' => '+', '_' => '/')) . '==');
  }

  /**
   * @throws NutException
   */
  private function decode() {
    $ref = $this->nut_encoded;
    $len = strlen($ref);
    if ($len === 16 || $len === 28) {
      $this->nut_raw = array(
        'time'    => $this->decode_time($ref),
        'counter' => $this->decode_counter($ref),
        'random'  => $this->decode_random($ref),
        'ip'      => $this->decode_ip($ref),
      );
    }
    else {
      throw new NutException('Nut string length incorrect: ' . $len);
    }
  }

  /**
   * @param string $bytes
   * @return string
   */
  private function decode_time($bytes): string {
    $output = unpack('L', $this->_bytes_extract($bytes, 0, 4));
    return array_shift($output);
  }

  /**
   * @param string $bytes
   * @return string
   */
  private function decode_counter($bytes): string {
    $output = unpack('L', $this->_bytes_extract($bytes, 4, 4));
    return array_shift($output);
  }

  /**
   * @param string $bytes
   * @return string
   */
  private function decode_random($bytes): string {
    return $this->_bytes_extract($bytes, 8, 4);
  }

  /**
   * @param string $bytes
   *
   * @return string
   * @throws \Drupal\sqrl\Exception\NutException
   */
  private function decode_ip($bytes): string {
    $length = (strlen($bytes) === 16) ? 4 : 16;
    return $this->_dtr_ntop($this->_bytes_extract($bytes, 12, $length));
  }

  /**
   *
   */
  private function validate_expiration() {
    $this->expired = FALSE;
    if ($this->nut_raw['time'] + State::EXPIRE_NUT > $this->time->getRequestTime()) {
      $this->expired = TRUE;
      // TODO: Delete cookie.
    }
  }

  /**
   * @throws NutException
   */
  private function load() {
    $params = $this->state->getNut($this->nut_public);
    if (empty($params)) {
      throw new NutException('No params received from implementing framework');
    }
    if (empty($params['op'])) {
      throw new NutException('Wrong params received from implementing framework');
    }
    if (empty($params['ip'])) {
      throw new NutException('Wrong params received from implementing framework');
    }
    if (!isset($params['params']) || !is_array($params['params'])) {
      throw new NutException('Wrong params received from implementing framework');
    }
    if (!isset($params['messages to browser']) || !is_array($params['messages to browser'])) {
      throw new NutException('Wrong params received from implementing framework');
    }
    if (!isset($params['uid'])) {
      throw new NutException('Wrong params received from implementing framework');
    }

    $this->clientOperation = $params['op'];
    $this->clientOperationParams = $params['params'];
    $this->clientMessages = $params['messages to browser'];
    $this->clientUid = $params['uid'];
  }

}
