<?php

namespace Drupal\sqrl\Plugin\SqrlAction;

use Drupal\Component\Plugin\Exception\PluginException;
use Drupal\sqrl\SqrlActionPluginBase;
use Drupal\user\Entity\User;
use Drupal\user\UserInterface;
use Exception;

/**
 * Plugin implementation of the sqrl action "login".
 *
 * @SqrlAction(
 *   id = "login",
 *   label = @Translation("Login / register"),
 *   description = @Translation("TBD.")
 * )
 */
class Login extends SqrlActionPluginBase {

  /**
   * @var int
   */
  private $tif;

  /**
   * {@inheritdoc}
   */
  public function run(): bool {
    $this->tif = $this->client->getTif();
    switch ($this->client->getNut()->getClientOperation()) {
      case 'register':
        return $this->register();
        break;

      case 'login':
        return $this->login();
        break;

      case 'link':
        return $this->link();
        break;

    }

    // TODO: Error handling.
    return FALSE;
  }

  /**
   * @return bool
   */
  private function login(): bool {
    if (!$this->user) {
      return $this->register();
    }

    if (!$this->user->isActive()) {
      return FALSE;
    }

    if (($this->tif & $this->client::FLAG_PIDK_MATCH) === $this->client::FLAG_PIDK_MATCH) {
      // The login request got authenticated with the PIDK. We do respond with
      // the SUK and VUK from the currently known ID and expect the client then
      // to issue a setkey request with the new SUK and VUK matching the new IDK.
      if ($identity = $this->getIdentityByPidk()) {
        $this->client->setResponse('suk', $identity->getSuk());
        $this->client->setResponse('vuk', $identity->getVuk());
      }
    }
    else {
      $this->client->setMessage($this->t('Login successful!'));
      return TRUE;
    }
    return FALSE;
  }

  /**
   * @return bool
   */
  private function register(): bool {
    if ($this->user) {
      return $this->login();
    }

    $registerMethod = $this->configFactory->get('user.settings')->get('register');
    if ($registerMethod === UserInterface::REGISTER_ADMINISTRATORS_ONLY) {
      // We can't create new accounts on this server.
      $this->log->debug('No user registration allowed!');
      $this->client->setMessage($this->t('No user registration allowed!'));
      return FALSE;
    }

    $userinfo = array(
      'name' => '',
      'pass' => $this->random->string(32),
      'init' => '',
      'status' => $registerMethod === UserInterface::REGISTER_VISITORS,
      'access' => $this->time->getRequestTime(),
    );
    $pattern = $this->configFactory->get('sqrl.settings')->get('account_name_pattern');
    if (strpos($pattern, '[rand]') === FALSE) {
      $pattern .= '[rand]';
    }
    $length = $this->configFactory->get('sqrl.settings')->get('account_name_random_length');
    while (empty($user)) {
      $name = str_replace('[rand]', $this->random->string($length), $pattern);
      if (!user_load_by_name($name)) {
        $userinfo['name'] = $userinfo['init'] = $name;
        $userinfo['mail'] = $this->dummyMail($name);
        try {
          $user = User::create($userinfo);
          $user->save();
        }
        catch (Exception $ex) {
          // Do nothing, just retry.
          unset($user);
        }
      }
    }
    $this->user = $user ?? NULL;
    if ($this->link()) {
      if ($registerMethod === UserInterface::REGISTER_VISITORS_ADMINISTRATIVE_APPROVAL) {
        // New accounts require admin approval.
        $this->client->setMessage($this->t('Account created, please wait for an administrator for approval of your account.'));
      }
      return TRUE;
    }
    return FALSE;
  }

  /**
   * @return bool
   */
  private function link(): bool {
    if (!$this->user) {
      return FALSE;
    }
    try {
      $plugin = $this->actionPluginManager->createInstance('setkey');
      $plugin->setUser($this->user);
      if ($plugin->run()) {
        return TRUE;
      }
    } catch (PluginException $e) {
    }
    return FALSE;
  }

  /**
   * @param string $name
   *
   * @return string
   */
  private function dummyMail($name): string {
    return urlencode($name) . '@local.localhost';
  }

}
