<?php

namespace Drupal\sqrl;

use Drupal\sqrl\Exception\NutException;
use Exception;

trait StringManipulation {

  /**
   * return number of microseconds since last unix timestamp change,
   * scaled  by 10 modulo 65536, see Apache Module mod_unique_id for
   * justification that this is adequately request unique.
   *
   * @return int
   */
  protected function counter(): int {
    list($usec, ) = explode(' ', microtime());
    return ($usec * 100000) % 65536;
  }

  /**
   * dtr_pton
   *
   * Converts a printable IP into an unpacked binary string
   *
   * @param string $ip
   *
   * @return string
   * @throws \Drupal\sqrl\Exception\NutException
   * @author Mike Mackintosh - mike@bakeryphp.com
   */
  protected function _dtr_pton($ip): string {
    if (filter_var($ip, FILTER_VALIDATE_IP, FILTER_FLAG_IPV4)) {
      return current(unpack('A4', inet_pton($ip)));
    }

    if (filter_var($ip, FILTER_VALIDATE_IP, FILTER_FLAG_IPV6)) {
      return current(unpack('A16', inet_pton($ip)));
    }

    throw new NutException('Please supply a valid IPv4 or IPv6 address');
  }

  /**
   * dtr_ntop
   *
   * Converts an unpacked binary string into a printable IP
   *
   * @param string $str
   *
   * @return string
   * @throws \Drupal\sqrl\Exception\NutException
   * @author Mike Mackintosh - mike@bakeryphp.com
   */
  protected function _dtr_ntop($str): string {
    $len = strlen($str);
    if ($len === 16 || $len === 4) {
      return inet_ntop(pack('A'. strlen($str), $str));
    }
    throw new NutException('Please provide a 4 or 16 byte string');
  }

  /**
   * @param int $length
   *   Length of the password to be generated.
   *
   * @return string
   */
  protected function randomBytes($length): string {
    $randomBytes = '';
    try {
      $randomBytes = random_bytes($length);
    }
    catch (Exception $e) {
      // TODO: Error handling.
    }
    return $randomBytes;
  }

  /**
   * Returns a URL safe base64 encoded version of the string.
   *
   * @param string $string
   * @return string
   */
  protected function base64_encode($string): string {
    $data = base64_encode($string);
    // Modify the output so it's safe to use in URLs.
    return strtr($data, array('+' => '-', '/' => '_', '=' => ''));
  }

  /**
   * Returns the base64 decoded version of the URL safe string.
   *
   * @param string $string
   * @return string
   */
  protected function base64_decode($string): string {
    $string = strtr($string, array('-' => '+', '_' => '/'));
    return base64_decode($string);
  }

  /**
   * Extract specific bytes from a byteString
   *
   * @param string $bytes
   *  STRING of bytes
   * @param int $start
   *  INT string pointer to first byte to be extracted
   * @param int $len
   *  INT number of butes to extract
   * @return string
   *  STRING of bytes
   */
  protected function _bytes_extract($bytes, $start, $len): string {
    $result = '';
    while ($len > 0) {
      $result .= $bytes[$start];
      $start++;
      $len--;
    }
    return $result;
  }

}
